// MultiStepper.cpp
//
// Copyright (C) 2015 Mike McCauley
// $Id:  $

#include "NodeManager.h"
#include "Stepnode.h"

NodeManager::NodeManager(HardwareSerial* serial, uint8_t n) {
    _serial = serial;
    _n = n;

    // create nodes
    _nodes = (Stepnode *) malloc(_n * sizeof(Stepnode));
    uint8_t i;
    for (i = 0; i < _n; i++) {
        _nodes[i] = Stepnode(this, i);
    }
}

void NodeManager::init() {
    _serial->begin(NODE_BAUDRATE);
    _serial->setTimeout(NODE_TIMEOUT_MS);

    // flush serial
    flush_input();

    // initial sync
    send = 0;
    recv = 0;
    sync_both();
}

void NodeManager::setEnabled(uint8_t i, bool enabled) {
    if (_nodes[i].state_enabled != enabled) {
        send = 0b11 << (2*i);
        sync_both();
    }
}

bool NodeManager::getEnabled(uint8_t i) {
    return _nodes[i].state_enabled;
}

bool NodeManager::getEndstop(uint8_t i) {
    return _nodes[i].state_endstop;
}

void NodeManager::setMaxSpeed(uint8_t i, float speed) {
    _nodes[i].setMaxSpeed(speed);
}

void NodeManager::setAcceleration(uint8_t i, float acc) {
    _nodes[i].setAcceleration(acc);
}

void NodeManager::setCurrentPosition(uint8_t i, long pos) {
    _nodes[i].setCurrentPosition(pos);
}

long NodeManager::getCurrentPosition(uint8_t i) {
    return _nodes[i].currentPosition();
}

void NodeManager::setTarget(uint8_t i, long pos) {
    return _nodes[i].moveTo(pos);
}

void NodeManager::runToTarget() {
    while (run());

    flush_input();
    // sync state
    send = 0;
    sync_both();
}

bool NodeManager::run() {
    uint8_t i;
    bool running = false;
    send = 0;
    for (i = 0; i < _n; i++) {
        running |= _nodes[i].run();
    }
    if (send != 0) {
        sync_send();
    }
    return running;
}

bool NodeManager::homing(uint8_t i, long pos_max, long pos_new, float speed) {
    float speed_old = _nodes[i].maxSpeed();
    _nodes[i].setMaxSpeed(speed);
    _nodes[i].moveTo(pos_max);

    bool touched = _nodes[i].state_endstop;
    bool running = true;
    while (running && _nodes[i].state_endstop == ENDSTOP_OPEN) {
        send = 0;
        running = _nodes[i].run();
        sync_both();
    }
    _nodes[i].setMaxSpeed(speed_old);

    // failure
    if (_nodes[i].state_endstop == ENDSTOP_OPEN) {
        return false;
    }
    _nodes[i].setCurrentPosition(pos_new);
    return true;
}

void NodeManager::flush_input() {
    char c;
    while (_serial->readBytes(&c, 1) != 0);
}

inline void NodeManager::sync_send() {
    _serial->write(send);
}

bool NodeManager::sync_both() {
    _serial->write(send);
    if (_serial->readBytes(&recv,1) == 0) {
        return false;
    }
    uint8_t i;
    // shifted version
    char c = recv >> (2*(4-_n));
    for (i = 0; i < _n; i++) {
        _nodes[i].state_enabled = (c & (0b10 << (2*i))) != 0;
        _nodes[i].state_endstop = (c & (0b01 << (2*i))) != 0;
    }
    return true;
}

NodeManager::~NodeManager() {
    free(_nodes);
}
